import os
import payload as pl
import pytest


class Fixtures(object):
    @pytest.fixture
    def api_key(self):
        print('API_KEY')
        pl.api_key = os.environ["TEST_SECRET_KEY"]
        if "TEST_API_URL" in os.environ:
            pl.api_url = os.environ["TEST_API_URL"]

    @pytest.fixture
    def customer_account(self):
        customer_account = pl.Customer.create(name="Test", email="test@example.com")
        return customer_account

    @pytest.fixture
    def processing_account(self):
        processing_account = pl.ProcessingAccount.create(
            {
                "name": "Processing Account",
                "legal_entity": {
                    "legal_name": "Test",
                    "type": "INDIVIDUAL_SOLE_PROPRIETORSHIP",
                    "ein": "23 423 4234",
                    "street_address": "123 Example Street",
                    "unit_number": "Suite 1",
                    "city": "New York",
                    "state_province": "NY",
                    "state_incorporated": "NY",
                    "postal_code": "11238",
                    "phone_number": "(111) 222-3333",
                    "website": "www.payload.co",
                    "start_date": "05/01/2015",
                    "contact_name": "Test Person",
                    "contact_email": "test.person@example.com",
                    "contact_title": "VP",
                    "owners": [
                        {
                            "full_name": "Test Person",
                            "email": "test.person@example.com",
                            "ssn": "234 23 4234",
                            "birth_date": "06/20/1985",
                            "title": "CEO",
                            "ownership": "100",
                            "street_address": "123 Main Street",
                            "unit_number": "#1A",
                            "city": "New York",
                            "state_province": "NY",
                            "postal_code": "10001",
                            "phone_number": "(111) 222-3333",
                            "type": "owner",
                        }
                    ],
                },
                "payment_methods": {
                    "type": "bank_account",
                    "bank_account": {
                        "account_number": "123456789",
                        "routing_number": "036001808",
                        "account_type": "checking",
                    },
                },
            }
        )
        return processing_account

    @pytest.fixture
    def card_payment(self, processing_account):
        card_payment = pl.Payment.create(
            processing_id=processing_account.id,
            amount=100.0, payment_method=pl.Card(card_number="4242 4242 4242 4242", expiry="12/25")
        )
        return card_payment

    @pytest.fixture
    def bank_payment(self, processing_account):

        bank_payment = pl.Payment.create(
            type="payment",
            amount=100.0,
            payment_method=pl.BankAccount(
                account_number="1234567890",
                routing_number="036001808",
                account_type="checking",
            ),
        )

        return bank_payment



import json

class PayloadError(Exception):
    http_code = None
    details = None
    response = None
    def __init__( self, description=None, response=None ):
        if not description:
            description = self.__class__.__name__
        if response:
            self.response = response
            self.details = self.response.get('details')
            description += '\n\n'+json.dumps(
                response,
                sort_keys=True,
                indent=4
            )
        super(PayloadError, self).__init__( description )

class UnknownResponse(PayloadError):
    pass

class BadRequest(PayloadError):
    http_code = 400

class TransactionDeclined(BadRequest):
    def __init__(self, *args, **kwargs):
        from .utils import map_object
        super().__init__(*args, **kwargs)
        self.transaction = self.details = map_object(self.response['details'])

class InvalidAttributes(BadRequest):
    pass

class Unauthorized(PayloadError):
    http_code = 401

class Forbidden(PayloadError):
    http_code = 403

class NotFound(PayloadError):
    http_code = 404

class TooManyRequests(PayloadError):
    http_code = 429

class InternalServerError(PayloadError):
    http_code = 500

class ServiceUnavailable(PayloadError):
    http_code = 503



import pytest
import os
import string
import random

import payload
from payload.exceptions import BadRequest, NotFound

from .fixtures import Fixtures

import dateutil.parser

@pytest.fixture
def pl():
    payload.api_key = None
    pl = payload.Session(os.environ["TEST_SECRET_KEY"])
    if "TEST_API_URL" in os.environ:
        pl.api_url = os.environ["TEST_API_URL"]

    return pl

class TestSession(Fixtures):
    def test_create_customer_account(self, pl):
        customer_account = pl.Customer.create(name="Test", email="test@example.com")
        assert customer_account.id

    def test_delete(self, pl):
        customer_account = pl.Customer.create(name="Test", email="test@example.com")
        with pytest.raises(NotFound):
            customer_account.delete()
            customer_get = pl.Customer.get(customer_account.id)

    def test_create_mult_accounts(self, pl):
        letters = string.ascii_lowercase
        rand_email1 = "".join(random.choice(letters) for i in range(5)) + "@example.com"
        rand_email2 = "".join(random.choice(letters) for i in range(5)) + "@example.com"

        accounts = pl.create(
            [
                pl.Customer(email=rand_email1, name="Matt Perez"),
                pl.Customer(email=rand_email2, name="Andrea Kearney"),
            ]
        )

        get_account_1 = pl.Customer.filter_by(email=rand_email1).all()[0]
        get_account_2 = pl.Customer.filter_by(email=rand_email2).all()[0]

        assert get_account_1
        assert get_account_2

    #def test_get_processing_account(self, api_key, processing_account):
    #    assert pl.ProcessingAccount.get(processing_account.id)
    #    assert processing_account.status == "pending"

    def test_paging_and_ordering_results(self, pl):
        accounts = pl.create(
            [
                pl.Customer(email="account1@example.com", name="Randy Robson"),
                pl.Customer(email="account2@example.com", name="Brandy Bobson"),
                pl.Customer(email="account3@example.com", name="Mandy Johnson"),
            ]
        )

        customers = pl.Customer.filter_by(
            order_by="created_at", limit=3, offset=1
        ).all()

        assert len(customers) == 3
        assert dateutil.parser.parse(customers[0].created_at) < dateutil.parser.parse(customers[1].created_at)
        assert dateutil.parser.parse(customers[1].created_at) < dateutil.parser.parse(customers[2].created_at)

    def test_update_cust(self, pl):
        customer_account = pl.Customer.create(name="Test", email="test@example.com")
        customer_account.update(email="test2@example.com")

        assert customer_account.email == "test2@example.com"

    def test_update_mult_acc(self, pl):
        customer_account_1 = pl.Customer.create(
            name="Brandy", email="test1@example.com"
        )
        customer_account_2 = pl.Customer.create(name="Sandy", email="test2@example.com")

        pl.update(
            [
                [customer_account_1, {"email": "brandy@example.com"}],
                [customer_account_2, {"email": "sandy@example.com"}],
            ]
        )

        assert customer_account_1.email == "brandy@example.com"
        assert customer_account_2.email == "sandy@example.com"

    def test_get_cust(self, pl):
        customer_account = pl.Customer.create(name="Test", email="test@example.com")
        assert pl.Customer.get(customer_account.id)

    def test_select_cust_attr(self, pl):
        customer_account = pl.Customer.create(name="Test", email="test@example.com")
        select_customer_id = pl.Customer.select("id").get(customer_account.id).get("id")
        assert select_customer_id == customer_account.id



from six import with_metaclass
import operator

class Filter(object):

    def __init__(self, attr, val):
        self.attr  = str(attr)
        self.val   = val
        self.opval = self.op+str(val)

    def __or__(self, other):
        if not isinstance( other, Filter ):
            raise TypeError('invalid type')
        if other.attr != self.attr:
            raise ValueError('`or` only works on the same attribute')
        return Filter( self.attr, '|'.join(
            self.opval.split('|') + other.opval.split('|') ) )

class Equal(Filter):
    op = ''

class NotEqual(Filter):
    op = '!'

class GreaterThan(Filter):
    op = '>'

class LessThan(Filter):
    op = '<'

class GreaterThanEqual(Filter):
    op = '>='

class LessThanEqual(Filter):
    op = '<='

class Contains(Filter):
    op = '?*'

class MetaAttr(type):
    def __getattr__(cls, key):
        return Attr(key)
    def __iter__(cls):
        yield '*'

class Attr(with_metaclass(MetaAttr)):
    is_method = False

    def __init__(self, param, parent=None):
        self.param = param
        self.parent = parent
        if not parent:
            self.key = self.param
        else:
            self.key = '{}[{}]'.format( self.parent.key, self.param )

    def __getattr__(self, key):
        if self.is_method:
            raise ValueError('cannot get attr of method')
        return Attr(key,parent=self)

    def __str__(self):
        if self.is_method:
            return '{}({})'.format( self.param, self.parent.key )
        return self.key

    def __eq__(self, other):
        return Equal( self, other )

    def __ne__(self, other):
        return NotEqual( self, other )

    def __gt__(self, other):
        return GreaterThan( self, other )

    def __lt__(self, other):
        return LessThan( self, other )

    def __ge__(self, other):
        return GreaterThanEqual( self, other )

    def __le__(self, other):
        return LessThanEqual( self, other )

    def contains(self, other):
        return Contains( self, other )

    def __call__(self):
        self.is_method = True
        return self



"""
Payload python library
------------------------

Documentation: https://docs.payload.co

:copyright: (c) 2021 Payload (http://payload.co)
:license: MIT License
"""

from .version import __version__
from .exceptions import *
from .objects import *
from .arm import Attr as attr, ARMRequest, session_factory
from . import objects

URL = 'https://api.payload.co'

api_key = None
api_url = URL

Session = session_factory('PayloadSession', objects)

def create(*args, **kwargs):
	return ARMRequest().create(*args, **kwargs)

def update(*args, **kwargs):
	return ARMRequest().update(*args, **kwargs)

def delete(*args, **kwargs):
	return ARMRequest().delete(*args, **kwargs)



from .attr import MetaAttr
from functools import partial
from six import with_metaclass
import json

_object_cache = {}

class ARMMetaObject(MetaAttr):
    def __new__(cls, name, bases, dct):
        if len(bases) and hasattr( bases[0], '__spec__' ):
            dct['__spec__'] = dict(bases[0].__spec__, **dct['__spec__'])

        self = super(ARMMetaObject, cls).__new__(cls, name, bases, dct)

        if len(bases):
            _object_cache[self] = {}
            if 'endpoint' not in self.__spec__:
                self.__spec__['endpoint'] = '/{}s'.format(self.__spec__['object'])

        return self

class ARMObject(with_metaclass(ARMMetaObject)):
    __spec__ = {}
    _session = None
    field_map = set()

    def __new__(cls, _session=None, **obj):
        if 'id' in obj:
            uid = obj['id']
            if uid in _object_cache[cls]:
                _object_cache[cls][uid]._set_data(obj)
                return _object_cache[cls][uid]
        return super(ARMObject, cls).__new__(cls)

    def __init__(self, _session=None, **obj):
        self._session = _session
        self._set_data(dict(self.__spec__.get('polymorphic') or {}, **obj))

    def __getattr__(self, attr):
        if attr in self.field_map:
            return self._data.get(self._data.get('type'),{}).get(attr)
        return self._data.get(attr)

    def _set_data(self, obj):
        self._data = obj
        self._data = data2object(self._data, self.field_map, self._session)
        if 'id' in obj:
            _object_cache[self.__class__][obj['id']] = self

    def json(self):
        return json.dumps( self.data(), indent=4, sort_keys=True )

    def data(self):
        return object2data(self._data)

    def update(self, **update):
        ARMRequest(self.__class__, self._session)._request('put', id=self.id, json=update)

    def delete(self):
        ARMRequest(self.__class__, self._session)._request('delete', id=self.id)

    @classmethod
    def get(cls, id, *args, **kwargs):
        return ARMRequest(cls, kwargs.pop('_session', None)).get(id, *args, **kwargs)

    @classmethod
    def filter_by(cls, *filters, **kw_filters):
        return ARMRequest(cls, kw_filters.pop('_session', None))\
            .filter_by(*filters, **dict(cls.__spec__.get('polymorphic') or {}, **kw_filters))

    @classmethod
    def create(cls, objects=None, **values):
        return ARMRequest(cls, values.pop('_session', None))\
            .create(objects, **values)

    @classmethod
    def select(cls, *fields, _session=None):
        return ARMRequest(cls, _session).select(*fields)\
            .filter_by(**dict(cls.__spec__.get('polymorphic') or {}))

    @classmethod
    def update_all(cls, objects=None, **values):
        return ARMRequest(cls, values.pop('_session', None))\
            .update(objects, **values)

    @classmethod
    def delete_all(cls, objects=None, **values):
        return ARMRequest(cls, values.pop('_session', None))\
            .delete(objects, **values)

class ARMMetaObjectWrapper(object):
    def __getattr__(cls, key):
        if key == 'Foo':
            return cls._foo_func()
        elif key == 'Bar':
            return cls._bar_func()
        raise AttributeError(key)

class ARMObjectWrapper(object):
    def __init__(self, Object, session):
        self.Object = Object
        self.session = session

    def __call__(self, *args, **kwargs):
        obj = self.Object(*args, **kwargs)
        obj._session = self.session
        return obj

    def call(self, __name, *args, **kwargs):
        kwargs['_session'] = self.session
        return getattr(self.Object, __name)(*args, **kwargs)

    def __getattr__(self, name):
        if name in ('get', 'filter_by', 'create', 'select', 'update_all', 'delete_all'):
            return partial(self.call, name)

        return getattr(self.Object, name)


from .request import ARMRequest
from ..utils import object2data, data2object



import pytest

import payload as pl

from .fixtures import Fixtures


@pytest.fixture()
def billing_schedule(processing_account, customer_account):
    billing_schedule = pl.BillingSchedule.create(
        start_date="2019-01-01",
        end_date="2019-12-31",
        recurring_frequency="monthly",
        type="subscription",
        customer_id=customer_account.id,
        processing_id=processing_account.id,
        charges=pl.BillingCharge(type="option_1", amount=39.99),
    )

    return billing_schedule


class TestBilling(Fixtures):
    def test_create_billing_schedule(
        self, api_key, billing_schedule, processing_account
    ):
        assert billing_schedule.processing_id == processing_account.id
        assert billing_schedule.charges[0].amount == 39.99

    def test_update_billing_schedule_frequency(
        self, api_key, billing_schedule, processing_account
    ):
        assert billing_schedule.processing_id == processing_account.id
        assert billing_schedule.charges[0].amount == 39.99

        billing_schedule.update(recurring_frequency="quarterly")

        assert billing_schedule.recurring_frequency == "quarterly"



import pytest

import payload as pl

from .fixtures import Fixtures


class TestPaymentLink(Fixtures):
    def test_create_payment_link(self, api_key, processing_account):
        payment_link = pl.PaymentLink.create(
            type="one_time",
            description="Payment Request",
            amount=10.00,
            processing_id=processing_account.id,
        )

        assert payment_link.processing_id == processing_account.id



import pytest

import payload as pl
from payload.exceptions import NotFound

from .fixtures import Fixtures


class TestTransaction(Fixtures):
    def test_transaction_ledger_empty(self, api_key, card_payment):
        transaction = pl.Transaction.select("*", "ledger").get(card_payment.id)
        assert transaction.ledger == []

    def test_unified_payout_batching(self, api_key, processing_account):
        pl.Refund.create(
            amount=10,
            processing_id=processing_account.id,
            payment_method=pl.Card(card_number="4242 4242 4242 4242", expiry="12/25"),
        )

        transactions = (
            pl.Transaction.select("*", "ledger")
            .filter_by(type="refund", processing_id=processing_account.id)
            .all()
        )

        assert len(transactions) == 1
        assert transactions[0].processing_id == processing_account.id

    def test_get_transactions(self, api_key, card_payment):
        payments = pl.Transaction.filter_by(status="processed", type="payment").all()
        assert len(payments) > 0

    def test_risk_flag(self, api_key, card_payment):
        assert card_payment.risk_flag == "allowed"

    def test_update_processed(self, api_key, card_payment):
        card_payment.update(status="voided")
        assert card_payment.status == "voided"

    def test_transactions_not_found(self, api_key, card_payment):
        with pytest.raises(NotFound):
            pl.Transaction.get("invalid")



import random
import string

import pytest

import payload as pl
from payload.exceptions import BadRequest

from .fixtures import Fixtures


class TestPaymentMethod(Fixtures):
    def test_create_payment_card(self, api_key, card_payment):
        assert card_payment.status == "processed"

    def test_create_payment_bank(self, api_key, bank_payment):
        assert bank_payment.status == "processed"

    def test_payment_filters(self, api_key, processing_account):
        letters = string.ascii_lowercase
        rand_description = "".join(random.choice(letters) for i in range(10))

        card_payment = pl.Payment.create(
            amount=100.0,
            description=rand_description,
            processing_id=processing_account.id,
            payment_method=pl.Card(
                card_number="4242 4242 4242 4242", expiry="05/22", card_code="123"
            ),
        )

        payments = pl.Payment.filter_by(
            pl.attr.amount > 99,
            pl.attr.amount < (200),
            pl.attr.description.contains(rand_description),
            pl.attr.created_at > ("2019-12-31"),
        ).all()

        assert len(payments) == 1
        assert payments[0].id == card_payment.id

    def test_void_card_payment(self, api_key, card_payment):
        card_payment.update(status="voided")

        assert card_payment.status == "voided"

    def test_void_bank_payment(self, api_key, bank_payment):
        bank_payment.update(status="voided")

        assert bank_payment.status == "voided"

    def test_refund_card_payment(self, api_key, card_payment):

        refund = pl.Refund.create(
            amount=100, ledger=[pl.Ledger(assoc_transaction_id=card_payment.id)]
        )

        assert refund.type == "refund"
        assert refund.amount == 100
        assert refund.status_code == "approved"

    def test_partial_refund_card_payment(self, api_key, card_payment):

        refund = pl.Refund.create(
            amount=10, ledger=[pl.Ledger(assoc_transaction_id=card_payment.id)]
        )

        assert refund.type == "refund"
        assert refund.amount == 10
        assert refund.status_code == "approved"

    def test_blind_refund_card_payment(self, api_key, processing_account):
        refund = pl.Refund.create(
            amount=10,
            processing_id=processing_account.id,
            payment_method=pl.Card(card_number="4242 4242 4242 4242", expiry="12/25"),
        )

        assert refund.type == "refund"
        assert refund.amount == 10
        assert refund.status_code == "approved"

    def test_refund_bank_payment(self, api_key, bank_payment):
        refund = pl.Refund.create(
            amount=100, ledger=[pl.Ledger(assoc_transaction_id=bank_payment.id)]
        )

        assert refund.type == "refund"
        assert refund.amount == bank_payment.amount
        assert refund.status_code == "approved"

    def test_partial_refund_bank_payment(self, api_key, bank_payment):
        refund = pl.Refund.create(
            amount=10, ledger=[pl.Ledger(assoc_transaction_id=bank_payment.id)]
        )

        assert refund.type == "refund"
        assert refund.amount == 10
        assert refund.status_code == "approved"

    def test_invalid_payment_method_type_invalid_attributes(self, api_key):
        with pytest.raises(BadRequest):
            pl.Transaction.create(type="invalid", card_number="4242 4242 4242 4242", expiry="12/25")



from .arm.object import ARMObject, _object_cache
from .arm.attr import Attr

def get_object_cls(item):
    cls = None
    for Object in _object_cache:
        if Object.__spec__['object'] != item.get('object'):
            continue
        if 'polymorphic' in Object.__spec__:
            found = True
            for attr, val in Object.__spec__['polymorphic'].items():
                if item.get(attr) != val:
                    found = False
                    break
            if found:
                cls = Object
                break
        elif not cls:
            cls = Object
    return cls

def map_object(item):
    cls = get_object_cls(item)
    if cls:
        return cls(**item)
    return item

def nested_qstring_keys(base):
    def recurse(item, fmt='{}'):
        if isinstance( item, (list,tuple) ):
            _iter = enumerate(item)
        else:
            _iter = list(item.items())
        for key, val in _iter:
            new_key = fmt.format(key)
            if not isinstance(item[key], (dict, list, tuple)):
                if isinstance(val, Attr):
                    val = str(Attr)
                base[new_key] = val
            else:
                if item is base: del base[key]
                recurse(val, new_key+'[{}]')
        return base
    return recurse(base)

def data2object(item, field_map=set(), session=None):
    def recurse(item):
        if isinstance( item, (list,tuple) ):
            _iter = enumerate(item)
        else:
            convert_fieldmap(item, field_map)
            _iter = list(item.items())
        for key, val in _iter:
            if isinstance(val, (list, dict)):
                item[key] = recurse(val)
            if isinstance(val, dict)\
            and val.get('object'):
                Object = get_object_cls(val)
                if Object:
                    if session:
                        val['_session'] = session
                    item[key] = Object(**val)
        return item
    return recurse(item)

def convert_fieldmap(obj, field_map):
    for f in field_map:
        if 'type' not in obj: continue
        if obj['type'] not in obj:
            obj[obj['type']] = {}
        if f in obj:
            obj[obj['type']][f] = obj.pop(f)

def object2data(item):
    def recurse(item):
        if isinstance( item, (list,tuple) ):
            _iter = enumerate(item)
        else:
            if isinstance( item, ARMObject ):
                item = item.data()
            _iter = list(item.items())
        for key, val in _iter:
            if isinstance(val, ARMObject):
                item[key] = val.data()
            elif isinstance(val, (list, dict)):
                item[key] = recurse(val)
        return item
    return recurse(item)

def map_attrs( cls, item ):
    if isinstance( item, cls ):
        item = item.data()
    for key in item:
        if isinstance(item[key], ARMObject):
            map_attrs(item[key].__class__, item[key])
            continue
        if not cls.__spec__.get('attr_map')\
        or key not in cls.attr_map:
            continue
        key_map = cls.__spec__['attr_map'][key]
        if isinstance( new_key, str ):
            obj[key_map] = item.pop(key)
        else:
            nested_key = list(key_map.keys())[0]
            if nested_key not in obj:
                obj[nested_key] = {}
            obj[nested_key][key_map[nested_key]] = obj.pop(key)



from .request import ARMRequest
from .object import ARMObject, ARMObjectWrapper
from . import Attr
import types
import inspect
import payload

def get_object(objects, name):
    if isinstance(objects, dict):
        return objects[name]
    else:
        return getattr(objects, name)

class Session(object):
    attr = Attr

    def __init__(self, api_key=None, api_url=None):
        self.api_key = api_key or payload.api_key
        self.api_url = api_url or payload.api_url

    def create(self, *args, **kwargs):
        return ARMRequest(session=self).create(*args, **kwargs)

    def update(self, *args, **kwargs):
        return ARMRequest(session=self).update(*args, **kwargs)

    def delete(self, *args, **kwargs):
        return ARMRequest(session=self).delete(*args, **kwargs)

    def __getattr__(self, name):
        return ARMObjectWrapper(get_object(self._objects, name), self)

def session_factory(name, objects):
    return type(name, (Session,), {'_objects': objects})



import datetime

import pytest

import payload as pl

from .fixtures import Fixtures


@pytest.fixture
def invoice(processing_account, customer_account):
    invoice = pl.Invoice.create(
        type="bill",
        processing_id=processing_account.id,
        due_date=datetime.datetime.today().strftime('%Y-%m-%d'),
        customer_id=customer_account.id,
        items=[pl.ChargeItem(amount=29.99)],
    )

    return invoice


class TestInvoice(Fixtures):
    def test_create_invoice(self, api_key, invoice):
        assert invoice.due_date == datetime.datetime.today().strftime('%Y-%m-%d')
        assert invoice.status == "unpaid"

    def test_pay_invoice(self, api_key, invoice, customer_account):
        assert invoice.due_date == datetime.datetime.today().strftime('%Y-%m-%d')
        assert invoice.status == "unpaid"

        card_payment = pl.Card.create(
            account_id=customer_account.id, card_number="4242 4242 4242 4242", expiry='12/25'
        )

        if invoice.status != "paid":
            pl.Payment.create(
                amount=invoice.amount_due,
                customer_id=customer_account.id,
                payment_method_id=card_payment.id,
                allocations=[pl.PaymentItem(invoice_id=invoice.id)],
            )

        get_invoice = pl.Invoice.get(invoice.id)
        assert get_invoice.status == "paid"



from .arm.object import ARMObject

class Account(ARMObject):
    __spec__ = { 'object': 'account' }

class Customer(ARMObject):
    __spec__ = { 'object': 'customer' }

class ProcessingAccount(ARMObject):
    __spec__ = { 'object': 'processing_account' }

class Org(ARMObject):
    __spec__ = { 'endpoint': '/accounts/orgs', 'object': 'org' }

class User(ARMObject):
    __spec__ = { 'object': 'user' }

class Transaction(ARMObject):
    __spec__ = { 'endpoint': '/transactions', 'object': 'transaction' }

    def void(self):
        self.update(status='voided')
        return self

class Payment(Transaction):
    __spec__ = { 'polymorphic': { 'type': 'payment' } }

class Refund(Transaction):
    __spec__ = { 'polymorphic': { 'type': 'refund' } }

class Credit(Transaction):
    __spec__ = { 'polymorphic': { 'type': 'credit' } }

class Deposit(Transaction):
    __spec__ = { 'polymorphic': { 'type': 'deposit' } }

class Ledger(ARMObject):
    __spec__ = { 'object': 'transaction_ledger' }

class PaymentMethod(ARMObject):
    __spec__ = { 'object': 'payment_method' }

class Card(PaymentMethod):
    __spec__ = { 'polymorphic': { 'type': 'card' } }
    field_map = set(['card_number', 'expiry', 'card_code'])

class BankAccount(PaymentMethod):
    __spec__ = { 'polymorphic': { 'type': 'bank_account' } }
    field_map = set(['account_number', 'routing_number', 'account_type'])

class BillingSchedule(ARMObject):
    __spec__ = { 'object': 'billing_schedule' }

class BillingCharge(ARMObject):
    __spec__ = { 'object': 'billing_charge' }

class Invoice(ARMObject):
    __spec__ = { 'object': 'invoice' }

class LineItem(ARMObject):
    __spec__ = { 'object': 'line_item' }

class ChargeItem(LineItem):
    __spec__ = { 'polymorphic': { 'entry_type': 'charge' } }

class PaymentItem(LineItem):
    __spec__ = { 'polymorphic': { 'entry_type': 'payment' } }

class Webhook(ARMObject):
    __spec__ = { 'object': 'webhook' }

class PaymentLink(ARMObject):
    __spec__ = { 'object': 'payment_link' }

class PaymentActivation(ARMObject):
    __spec__ = { 'object': 'payment_activation' }




import os
import sys
try:
    from io import open
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

pkg_name = 'payload'

directory = os.path.abspath(os.path.dirname(__file__))

sys.path.insert(0, os.path.join(directory, pkg_name))

from version import __version__

with open(os.path.join(directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='payload-api',
    version='.'.join(map(str,__version__)),
    description='Payload Python Library',
    author='Payload',
    author_email='help@payload.co',
    url='https://github.com/payload-code/payload-python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=['requests', 'six'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



import os
import re
import sys

import payload
import requests
import copy

from ..utils import (convert_fieldmap, map_attrs, map_object,
                     nested_qstring_keys, object2data)
from .attr import Attr
from .object import ARMObject

if sys.version_info >= (3,0):
    from urllib.parse import urljoin
else:
    from urlparse import urljoin

class ARMRequest(object):

    def __init__(self, Object=None, session=None):
        self.Object = Object
        self.session = session
        self._filters  = []
        self._attrs    = []
        self._group_by = []

    def _request(self, method, id=None, headers=None, params=None, json=None):
        session  = self.session or payload
        endpoint = self.Object.__spec__['endpoint']
        headers  = headers or {}
        params   = nested_qstring_keys(params or {})
        auth     = (session.api_key, '')
        files    = {}

        if json:
            flat_data = nested_qstring_keys(copy.copy(json))
            for k in list(flat_data):
                if hasattr(flat_data[k], 'read'): files[k] = flat_data.pop(k)

        if id: endpoint = f"{endpoint}/{id}"

        if self._filters:
            for k, v in dict( (f.attr, f.opval) for f in self._filters ).items():
                if k not in params: params[k] = v

        if self._attrs:
            params['fields'] = list(map(str,self._attrs))

        if self._group_by:
            params['group_by'] = list(map(str,self._group_by))

        convert_fieldmap(params, self.Object.field_map)
        if json:
            convert_fieldmap(json, self.Object.field_map)
        params = nested_qstring_keys(params)

        if files:
            response = getattr(requests, method)(
                urljoin(session.api_url, endpoint.strip('/')),
                params=params,
                auth=auth,
                data=flat_data,
                files=files
                )
        else:
            response = getattr(requests, method)(
                urljoin(session.api_url, endpoint.strip('/')),
                headers=headers,
                params=params,
                auth=auth,
                json=json
                )
        try:
            data = response.json()

            if not isinstance(data, dict):
                raise payload.UnknownResponse()
        except ValueError:
            if response.status_code == 500:
                raise payload.InternalServerError()
            else:
                raise payload.UnknownResponse()

        if response.status_code == 200:
            if data.get('object') == 'list':
                for row in data['values']:
                    row['_session'] = self.session
                return list(map(map_object, data['values']))
            else:
                data['_session'] = self.session
                return map_object(data)
        else:
            errors = [ err for _ in payload.PayloadError.__subclasses__() for err in _.__subclasses__()+[_] ]
            for Error in errors:
                if Error.__name__ != data.get('error_type') \
                or Error.http_code != response.status_code:
                    continue
                raise Error(data.get('description'), data)
            if response.status_code == 500:
                raise payload.InternalServerError(data.get('description'), data)
            raise payload.BadRequest(data.get('description'), data)

    def get(self, id, **params):
        if not id:
            raise ValueError('id cannot be empty')
        return self._request('get', id=id, params=params)

    def select(self, *fields):
        self._attrs.extend(fields)
        return self

    def group_by(self, *fields):
        self._group_by.extend(fields)
        return self

    def create(self, obj=None, **values):
        obj = obj or values
        if isinstance( obj, list ):
            for o in obj:
                if not self.Object and not isinstance( o, ARMObject ):
                    raise TypeError('Bulk create requires ARMObject object types')
                if not self.Object:
                    self.Object = o.__class__
                elif isinstance( o, dict ):
                    o.update(self.Object.__spec__.get('polymorphic',{}))
                elif not isinstance( o, self.Object ):
                    raise TypeError('Bulk create requires all objects to be of the same type')
            obj = { 'object': 'list', 'values': obj }
        elif isinstance( obj, dict ):
            obj.update(self.Object.__spec__.get('polymorphic',{}))

        obj = object2data(obj)
        print(obj)
        return self._request('post', json=obj)

    def delete(self, objects=None):
        if isinstance( objects, list ):
            if not objects:
                raise ValueError('List must not be empty')
            for o in objects:
                if not isinstance( o, ARMObject ):
                    raise TypeError('Bulk create requires ARMObject object types')
                if not self.Object:
                    self.Object = o.__class__
                elif not isinstance( o, self.Object ):
                    raise TypeError('Bulk create requires all objects to be of the same type')
            delete_query = '|'.join([ obj.id for obj in objects ])
            return self._request('delete', params={'id': delete_query, 'mode': 'query'})
        elif isinstance( objects, ARMObject ):
            self.Object = objects.__class__
            return cls._request('delete', id=objects.id)
        elif objects is None and self.Object and self._filters:
            return self._request('delete', params={'mode':'query'})
        else:
            raise TypeError('Bulk create requires ARMObject object types')

    def update(self, objects=None, **values):
        if objects:
            if not isinstance( objects, list ):
                raise ValueError('first parameter must be a list of updates')
            if not objects or not isinstance( objects[0], (list, tuple) ):
                raise ValueError('first parameter must be a list of updates')
            for o, upd in objects:
                if not isinstance( o, ARMObject ):
                    raise TypeError('Bulk update requires ARMObject object types')
                if not self.Object:
                    self.Object = o.__class__
                elif not isinstance( o, self.Object ):
                    raise TypeError('Bulk update requires all objects to be of the same type')
            updates = { 'object': 'list', 'values':
                list(map( lambda upd: dict( upd[1], id=upd[0].id ), objects )) }
            return self._request('put', json=updates)
        return self._request('put', params={'mode':'query'}, json=values)

    def filter_by(self, *filters, **kw_filters):
        self._filters.extend(filters)
        for key, val in nested_qstring_keys(kw_filters).items():
            self._filters.append( getattr( Attr, key ) == val )
        return self

    def all(self):
        return self._request('get')

    def first(self):
        resp = self._request('get', params=dict(limit=1))
        if resp: return resp[0]
